import unittest
from pyneval.metric.diadem_metric import diadem_metric
from pyneval.model.swc_node import SwcNode, SwcTree
from pyneval.metric.utils.config_utils import get_avg_radius
from pyneval.io.read_json import read_json


class DiademTotTest(unittest.TestCase):
    def test_1(self):
        testTree = SwcTree()
        goldTree = SwcTree()
        goldTree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\diadem\diadem3.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\diadem\diadem3.swc")

        # goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\\gold\\ExampleGoldStandard.swc")
        # testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\ExampleTest.swc")
        get_avg_radius(goldTree)

        score = diadem_metric(swc_test_tree=testTree,
                              swc_gold_tree=goldTree,
                              config=read_json("D:\gitProject\mine\PyNeval\config\diadem_metric.json"))
        print(score[2])
        self.assertEqual(score[2], 1.0)

    def test_2(self):
        testTree = SwcTree()
        goldTree = SwcTree()
        goldTree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\diadem\diadem4.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\diadem\diadem4.swc")

        # goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\\gold\\ExampleGoldStandard.swc")
        # testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\ExampleTest.swc")
        get_avg_radius(goldTree)

        score = diadem_metric(swc_test_tree=testTree,
                              swc_gold_tree=goldTree,
                              config=read_json("D:\gitProject\mine\PyNeval\config\diadem_metric.json"))
        print(score[2])
        self.assertEqual(score[2], 0.5)

    def test_3(self):
        testTree = SwcTree()
        goldTree = SwcTree()
        goldTree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\ExampleGoldStandard.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\ExampleTest.swc")

        # goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\\gold\\ExampleGoldStandard.swc")
        # testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\ExampleTest.swc")
        get_avg_radius(goldTree)

        score = diadem_metric(swc_test_tree=testTree,
                              swc_gold_tree=goldTree,
                              config=read_json("D:\gitProject\mine\PyNeval\config\diadem_metric.json"))
        print(score[2])
        self.assertEqual(score[2], 0.9564541213063764)

    def test_4(self):
        testTree = SwcTree()
        goldTree = SwcTree()
        goldTree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\diadem\diadem8.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\diadem\diadem8.swc")

        get_avg_radius(goldTree)

        score = diadem_metric(swc_test_tree=testTree,
                              swc_gold_tree=goldTree,
                              config=read_json("D:\gitProject\mine\PyNeval\config\diadem_metric.json"))
        print(score[2])
        self.assertEqual(score[2], 1.0)

    def test_5(self):
        testTree = SwcTree()
        goldTree = SwcTree()
        goldTree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\diadem\diadem7.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\diadem\diadem7.swc")

        get_avg_radius(goldTree)

        score = diadem_metric(swc_test_tree=testTree,
                              swc_gold_tree=goldTree,
                              config=read_json("D:\gitProject\mine\PyNeval\config\diadem_metric.json"))
        print(score[2])
        self.assertEqual(score[2], 0.5)

    def test_6(self):
        testTree = SwcTree()
        goldTree = SwcTree()
        goldTree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\ExampleGoldStandard.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\ExampleTest.swc")

        get_avg_radius(goldTree)

        score = diadem_metric(swc_test_tree=testTree,
                              swc_gold_tree=goldTree,
                              config=read_json("D:\gitProject\mine\PyNeval\config\diadem_metric.json"))
        print(score)


if __name__ == '__main__':
    unittest.main()



# class_2_label = {
#     'bear':0,
#     'bird':1,
#     'cat':2,
#     'dog':3,
#     'giraffe':4,
#     'horse':5,
#     'sheep':6,
#     'zebra':7}


# class_names = list(class_2_label.keys())

class  class2label():
    def __init__(self, *initial_data, **kwargs):
        for dictionary in initial_data:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])



import os

from pyneval.model.swc_node import SwcTree
from pyneval.model.euclidean_point import Line
from pyneval.metric.utils.edge_match_utils import \
    get_idedge_dict, get_edge_rtree, get_lca_length, get_nearby_edges, get_route_node, \
    cal_rad_threshold, cal_len_threshold
from pyneval.tools.re_sample import down_sample_swc_tree_command_line
from anytree import PreOrderIter
from pyneval.io import swc_writer
from pyneval.io.read_json import read_json
import math
import numpy as np


# make sure the match path does not contain a detected redundant edge
def is_self_root_clean(gold_swc_tree, gold_line_tuple_a, gold_line_tuple_b, vis_list):
    # if two foots lay on the same edge, pass
    if gold_line_tuple_a[0].get_id() == gold_line_tuple_b[0].get_id() and \
            gold_line_tuple_a[1].get_id() == gold_line_tuple_b[1].get_id():
        return vis_list[gold_line_tuple_a[0].get_id()] != 1

    lca_id = gold_swc_tree.get_lca(gold_line_tuple_a[0].get_id(), gold_line_tuple_b[0].get_id())
    if lca_id is None:
        return False

    # get nodes on the route, make sure no extra node
    route_list_a = get_route_node(gold_line_tuple_a[0], lca_id)
    route_list_b = get_route_node(gold_line_tuple_b[0], lca_id)

    route_list = route_list_a + route_list_b
    for node in route_list:
        if node.get_id() == lca_id and \
                gold_line_tuple_a[0].get_id() != lca_id and \
                gold_line_tuple_b[0].get_id() != lca_id:
            continue
        if vis_list[node.get_id()] == 1:
            return False
    return True


def get_ang(side_edge1, side_edge2, oppo_edge):
    c = (side_edge1**2 + side_edge2**2 - oppo_edge**2) / (2 * side_edge1 * side_edge2)
    c = round(c, 4)
    if not (-1.0 <= c <= 1.0):
        raise Exception("[Error: ]In file overlap_detect.py, input is not a legal triangle {} {} {}".format(
            side_edge1, side_edge2, oppo_edge
        ))
    return math.acos(c) * 180 / math.pi


def color_origin_tree(new_swc_tree, swc_tree):
    new_swc_list = [node for node in PreOrderIter(new_swc_tree.root())]
    id_node_map = {}

    for node in PreOrderIter(swc_tree.root()):
        id_node_map[node.get_id()] = node

    for node in new_swc_list:
        if node._type == 3:
            o_node = id_node_map[node.get_id()]
            while o_node.get_id() != node.parent.get_id():
                o_node._type = 3
                o_node = o_node.parent


# find overlap edges
def get_self_match_edges_e_fast(swc_tree=None,
                                rad_threshold=None, len_threshold=None, list_size=-1,
                                mode="not_self", DEBUG=False):
    idx3d = get_edge_rtree(swc_tree)
    id_edge_dict = get_idedge_dict(swc_tree)
    # sort the tree according the tree size
    roots = swc_tree.root().children
    id_rootdis_dict = {}
    node_list = []
    root_list = []
    for root in roots:
        r_list = [node for node in PreOrderIter(root)]
        root_list.append(tuple([root, len(r_list)]))
    root_list.sort(key=lambda x: x[1])

    for root in root_list:
        r_list = [node for node in PreOrderIter(root[0])]
        node_list += r_list

    if list_size == -1:
        list_size = len(node_list)

    for node in node_list:
        id_rootdis_dict[node.get_id()] = node.root_length

    vis_list = np.zeros(list_size + 10)
    for node in node_list:
        if node.is_virtual() or node.parent.is_virtual():
            continue

        parent = node.parent

        rad_threshold1, rad_threshold2 = cal_rad_threshold(rad_threshold, node.radius(), node.parent.radius())

        line_tuple_as = get_nearby_edges(idx3d, node, id_edge_dict, threshold=rad_threshold1, not_self=True)
        line_tuple_bs = get_nearby_edges(idx3d, parent, id_edge_dict, threshold=rad_threshold2, not_self=True)

        done = False
        for line_tuple_a_d in line_tuple_as:
            if done:
                break
            for line_tuple_b_d in line_tuple_bs:
                line_tuple_a, line_tuple_b = line_tuple_a_d[0], line_tuple_b_d[0]
                dis_a, dis_b = line_tuple_a_d[1], line_tuple_b_d[1]

                test_length = get_lca_length(swc_tree,
                                             line_tuple_a,
                                             line_tuple_b,
                                             Line(e_node_1=node.get_center(),
                                                  e_node_2=node.parent.get_center()),
                                             id_rootdis_dict)
                gold_length = node.distance(parent)
                len_threshold1 = cal_len_threshold(len_threshold, gold_length)

                if dis_a <= rad_threshold1 and dis_b <= rad_threshold2 and \
                        math.fabs(test_length - gold_length) < len_threshold1:
                    if mode == "not_self" and not is_self_root_clean(swc_tree, line_tuple_a, line_tuple_b, vis_list):
                        continue
                    if DEBUG:
                        print("node = {} test_length = {} gold_length = {}".format(
                            node.get_id(), test_length, gold_length
                        ))
                    node._type = 3
                    node.parent._type = 3
                    vis_list[node.get_id()] = 1
                    done = True
                    break


def delete_overlap_node(swc_tree):
    node_list = swc_tree.get_node_list()
    for node in node_list:
        if node._type == 3:
            for son in node.children:
                son.parent = swc_tree.root()
            swc_tree.remove_node(node.parent, node)
    swc_tree.node_list = [node for node in PreOrderIter(swc_tree.root())]


def overlap_clean(swc_tree, out_path, file_name, loc_config=None):
    '''
    :param swc_tree: input swc tree
    :param out_path: output fold path
    :param file_name: name your file as you want
    :param loc_config:  load config/overlap_clean.json
    :return:
    '''
    dis_threshold, length_threshold, ang_threshold = \
        loc_config["radius_threshold"], loc_config["length_threshold"], loc_config["ang_threshold"]

    new_swc_tree = down_sample_swc_tree_command_line(swc_tree, loc_config)
    new_swc_tree = down_sample_swc_tree_command_line(new_swc_tree, loc_config)

    new_swc_tree.get_lca_preprocess(swc_tree.size())
    swc_tree.get_lca_preprocess(swc_tree.size())

    get_self_match_edges_e_fast(swc_tree=new_swc_tree,
                                rad_threshold=dis_threshold,
                                len_threshold=length_threshold,
                                list_size=swc_tree.size(),
                                mode="not_self", DEBUG=False)
    color_origin_tree(new_swc_tree, swc_tree)
    swc_writer.swc_save(new_swc_tree, os.path.join(out_path, os.path.join('marked_data', file_name)))
    delete_overlap_node(new_swc_tree)
    swc_writer.swc_save(new_swc_tree, os.path.join(out_path, os.path.join('clean_data', file_name)))



"""Plot functions
"""
import os
import logging
from logging import Logger
import numpy as np
import matplotlib.pyplot as plt
#import ipywidgets as widgets
from IPython.display import display
from PIL import Image

from .Utils import joinPath, datestamp, timestamp, setDir
from . import ConvPlotTools
from . import LabelDefinition

figs = [] # Array to contain all figures, easy to close later
#_PLOT_DIR = './myplots'

def closeAllPlots(plt):
    '''closeAllPlots: Close all plots.
    '''
    plt.close('all')

def createNewAxes(row=1, col=1, **kwargs):

    fig, axes = plt.subplots(nrows=row, ncols=col, **kwargs)
    return fig, axes

def createNewSubplot(*args, **kwargs):
    '''createNewSubplot: Create New Sub Plot(including 3d plot).
    Returns:
        axes (2d or 3d matplotlib axes): The return value.
    '''

    axes = plt.subplot(*args, **kwargs)
    return axes

def plotMyFigure(img, cmap='Greys_r'):
    """Plot an numpy array image
       img: numpy array
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    im = ax.imshow(img,cmap=cmap)

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(im, cax=cbar_ax)

    return ax

def plotMyFigureWithLabeBtn(img, cmap='Greys_r'):
    """Plot an numpy array image
       img: numpy array
    """
    # Define save dir
    dest_dir = joinPath(LabelDefinition.PlotLabelDir, datestamp())
    setDir(dest_dir)

    # Display image
    fig = plt.figure()
    ax = fig.add_subplot(111)
    im = ax.imshow(img,cmap=cmap)

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(im, cax=cbar_ax)

    # Display widget
    img_label_choose = widgets.RadioButtons(
        options=LabelDefinition.label_class_names,
        value='none',
        description='图片类别:',
        disabled=False)

    # Define callback
    def onRadioBtn(label):
        if (img.shape==LabelDefinition.LabelPatchSizeRestrict):
            label_img_name = str(LabelDefinition.class_d_label[label['new']]) \
                +'#plot_label#none#none#none#' \
                + timestamp() \
                + '#none#.tif'
            save_path = joinPath(dest_dir, label_img_name)
            print('\r' + label['new'], ' ,saved in: ', save_path)
            Image.fromarray(np.asarray(img, np.uint8)).save(save_path)
        else:
            print('Cannot label, the image size is {0} not {1}'.format(
                img.shape,
                LabelDefinition.LabelPatchSizeRestrict))
        img_label_choose.close()
        plt.close(fig)

    img_label_choose.observe(onRadioBtn, names='value')
    display(img_label_choose)
    return ax

def plotMyFigureOnAxes(ax, img, cmap='Greys_r'):
    """Plot an numpy array image on a given Axes
       :ax: Matplotlib Axes object
       :img: numpy array
    """
    im = ax.imshow(img,cmap=cmap)
    fig = ax.figure

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(im, cax=cbar_ax)

    return ax


def plot_conv_weights(weights, name, channels_all=True):
    """
    Plots convolutional filters
    :param weights: numpy array of rank 4
    :param name: string, name of convolutional layer
    :param channels_all: boolean, optional
    :return: nothing, plots are saved on the disk
    """
    # make path to output folder
#    plot_dir = joinPath(_PLOT_DIR, 'conv_weights')
#    plot_dir = joinPath(plot_dir, name)

    # create directory if does not exist, otherwise empty it
#    ConvPlotTools.prepare_dir(plot_dir, empty=True)

    w_min = np.min(weights)
    w_max = np.max(weights)

    channels = [0]
    # make a list of channels if all are plotted
    if channels_all:
        channels = range(weights.shape[2])

    # get number of convolutional filters
    num_filters = weights.shape[3]

    # get number of grid rows and columns
    grid_r, grid_c = ConvPlotTools.get_grid_dim(num_filters)

    # create figure and axes
    fig, axes = plt.subplots(min([grid_r, grid_c]),
                             max([grid_r, grid_c]))
    fig.suptitle(name, fontsize=20)

    # iterate channels
    for channel in channels:
        # iterate filters inside every channel
        for l, ax in enumerate(axes.flat):
            # get a single filter
            img = weights[:, :, channel, l]
            # put it on the grid
            im = ax.imshow(img, vmin=w_min, vmax=w_max, interpolation='nearest', cmap='seismic')
            # remove any labels from the axes
            ax.set_xticks([])
            ax.set_yticks([])

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(im, cax=cbar_ax)

    # save figure
#    plt.savefig(joinPath(plot_dir, '{}-{}.png'.format(name.replace('/','_'), channel)), bbox_inches='tight')


def plot_conv_output(conv_img, name):
    """
    Makes plots of results of performing convolution
    :param conv_img: numpy array of rank 4
    :param name: string, name of convolutional layer
    :return: nothing, plots are saved on the disk
    """
    # make path to output folder
#    plot_dir = joinPath(_PLOT_DIR, 'conv_output')
#    plot_dir = joinPath(plot_dir, name)

    # create directory if does not exist, otherwise empty it
#    ConvPlotTools.prepare_dir(plot_dir, empty=True)

    w_min = np.min(conv_img)
    w_max = np.max(conv_img)

    # get number of convolutional filters
    num_filters = conv_img.shape[3]

    # get number of grid rows and columns
    grid_r, grid_c = ConvPlotTools.get_grid_dim(num_filters)

    # create figure and axes
    fig, axes = plt.subplots(min([grid_r, grid_c]),
                             max([grid_r, grid_c]))
    fig.suptitle(name, fontsize=20)

    # iterate filters
    for l, ax in enumerate(axes.flat):
        # get a single image
        img = conv_img[0, :, :,  l]
        # put it on the grid
        im = ax.imshow(img, vmin=w_min, vmax=w_max, interpolation='bicubic', cmap='Greys_r')
        # remove any labels from the axes
        ax.set_xticks([])
        ax.set_yticks([])

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(im, cax=cbar_ax)

    # save figure
#    plt.savefig(joinPath(plot_dir, '{}.png'.format(name)), bbox_inches='tight')

def isTiffFile(name):
    if not name: return False

    name_parts = name.split('.')
    if len(name_parts) != 2: return False

    str_cnt = name_parts[0]
    str_end = name_parts[1]
    if str_end != 'tif': return False

    try:
        int(str_cnt)
    except ValueError as e:
        print(e)
        return False
    else:
        return True


def readImageSequence(dir):
    img_stack = []
    try:
        if not os.path.exists(dir):
            raise IOError('image dir not exist')
        img_names = os.listdir(dir)
        img_names = list(filter(isTiffFile, img_names))
        img_names = sorted(img_names, key=lambda x:int(x.split('.')[0]))
        for img_name in img_names:
            img_path = joinPath(dir, img_name)
            try:
                im = Image.open(img_path)
                im = np.asarray(im)
                img_stack.append(im)
            except IOError as e:
                print(e)
        return np.asarray(img_stack)
    except IOError as e:
        print(e)
    finally:
        if len(img_stack)==0:
            print('no valid image sequence')
            return None
        else:
            return np.asarray(img_stack)



import numpy as np

from pyneval.io import read_json
from pyneval.io.swc_writer import swc_save
from pyneval.metric.utils.klib import TiffFile
from pyneval.metric.utils.tiff_utils import front_expend_step
from pyneval.model.swc_node import SwcTree

TIFF_PATH = "../../data/optimation/6656_test.tif"


def tiff_cut(tiff_data, LDF = (0, 0, 0), len = (64, 64, 64)):
    res = tiff_data[LDF[0]:LDF[0]+len[0], LDF[1]:LDF[1]+len[1], LDF[2]:LDF[2]+len[2]]
    return res
